from connection import connection
from connection import config_pool
from context import ContextManager

__author__ = 'Erick Almeida'
version = "0.1.13"
__doc__ = """

    pypgwrap - efficient PostgreSQL database wrapper
    -------------------------------------------

    The 'efficient' module provides a efficient wrapper over psycopg2 supporting a
    Python API for common sql functions, explicit and implicit transactions mechanism and
    connection pooling.

    This is not intended to provide ORM-like functionality, just to make it
    easier to interact with PostgreSQL from python code for simple use-cases
    and allow direct SQL access for more complex operations.

    The module wraps the excellent 'psycopg2' library and most of the
    functionality is provided by this behind the scenes, except for pooling.

    The module provides:

        * Simplified handling of connections/cursor
            * Connection pool, single or multithreaded (inherited by psycopg2.pool)
            * Cursor context handler
            * Context Manager for explicit transactions
        * Python API to wrap basic SQL functionality
            * Simple select,update,delete,join methods extending the cursor
              context handler (also available as stand-alone methods which
              create an implicit cursor for simple queries) (from pgwrap)
        * Query results as dict (using psycopg2.extras.DictCursor)
        * Callable prepared statements
        * Logging support

    Basic usage
    -----------

        >>> import pypgwrap
        >>> pypgwrap.config_pool(max_pool=10, pool_expiration=1, url='postgres://localhost/', pool_manager=SimpleConnectionPool)
        >>> db = pypgwrap.connection()
        >>> with db.cursor() as c:
        ...     c.query('select version()')
        [['PostgreSQL...']]
        >>> v = db.query_one('select version()')
        >>> v
        ['PostgreSQL...']
        >>> v.items()
        [('version', 'PostgreSQL...')]
        >>> v['version']
        'PostgreSQL...'


    Basic usage, with transaction
    -----------
        Init pool at application start:

            >>> import pypgwrap
            >>> pypgwrap.config_pool(max_pool=10, pool_expiration=1, url='postgres://localhost/', pool_manager=SimpleConnectionPool)

        Explicit transactions:

            >>> db = pypgwrap.connection()
            >>> db.create_table('t1', '''id SERIAL PRIMARY KEY,
                                       name TEXT NOT NULL,
                                       count INTEGER NOT NULL DEFAULT 0,
                                       active BOOLEAN NOT NULL DEFAULT true''')
            >>> id0 = db.insert('doctest_t1', {'name': 'name_one'}, returning='id')['id']
            >>> id1 = db.insert('doctest_t1', {'name': 'name_two'}, returning='id')['id']
            >>> db.commit()

        Implicity transactions:

            >>> with pypgwrap.connection() as db:
            >>>     db.create_table('t1', '''id SERIAL PRIMARY KEY,
                                          name TEXT NOT NULL,
                                          count INTEGER NOT NULL DEFAULT 0,
                                          active BOOLEAN NOT NULL DEFAULT true''')
            >>>     id0 = db.insert('doctest_t1', {'name': 'name_one'}, returning='id')['id']
            >>>     id1 = db.insert('doctest_t1', {'name': 'name_two'}, returning='id')['id']

        Distributed transactions:

            >>> import uuid
            >>> key = uuid.uuid4()

            >>> with pypgwrap.connection(key=key) as db:
            >>>     db.create_table('t1', '''id SERIAL PRIMARY KEY,
                                          name TEXT NOT NULL,
                                          count INTEGER NOT NULL DEFAULT 0,
                                          active BOOLEAN NOT NULL DEFAULT true''')
            >>>     id0 = db.insert('doctest_t1', {'name': 'name_one'}, returning='id')['id']
            >>>     id1 = db.insert('doctest_t1', {'name': 'name_two'}, returning='id')['id']

            >>> db2 = pypgwrap.connection(key=key)
            >>> id3 = db.insert('doctest_t1', {'name': 'name_three'}, returning='id')['id']
            >>> id4 = db.insert('doctest_t1', {'name': 'name_four'}, returning='id')['id']

            >>> db3 = pypgwrap.connection(key=key)
            >>> db3.commit()

        Distributed transactions, with ContextManager:

            >>> with pypgwrap.ContextManager() as context:

            >>>     with pypgwrap.connection(key=context.key) as db:
            >>>         db.create_table('t1', '''id SERIAL PRIMARY KEY,
                                              name TEXT NOT NULL,
                                              count INTEGER NOT NULL DEFAULT 0,
                                              active BOOLEAN NOT NULL DEFAULT true''')
            >>>         id0 = db.insert('doctest_t1', {'name': 'name_one'}, returning='id')['id']
            >>>         id1 = db.insert('doctest_t1', {'name': 'name_two'}, returning='id')['id']

            >>>     db2 = pypgwrap.connection(key)
            >>>     id3 = db.insert('doctest_t1', {'name': 'name_three'}, returning='id')['id']
            >>>     id4 = db.insert('doctest_t1', {'name': 'name_four'}, returning='id')['id']


    Connection
    ----------

    The config_pool need some parameters:
        - max_pool: Maximum of connections created and mainteined in memory
        - pool_expiration: Idle time (in minutes) for close and destroy memory connection
        - url: Url with connection parameters

    The intention of this method is to call at application start up, only!

    The connection class provides methods to return a cursor object or execute SQL queries
    directly (using an implicit cursor).

    The connection context provides the following basic methods:

        cursor          - create a new instance of cursor class
        commit          - Commit transaction (called implicitly on exiting context handler)
        rollback        - Rollback transaction

    Cursor
    ------

    The module provides a cursor context handler wrapping the psycopg2 cursor.

    The cursor object uses the psycopg2 'DictCursor' by default (which
    returns rows as a pseudo python dictionary) however this can be overridden
    by providing a 'cursor_factory' parameter to the constructor.

    >>> db = pypgwrap.connection()
    >>> with db.cursor() as c:
    ...     c.query('select version()')
    [['PostgreSQL...']]

    The cursor context provides the following basic methods:

        execute         - execute SQL query and return rowcount
        query           - execute SQL query and fetch results
        query_one       - execute SQL query and fetch first result
        query_dict      - execute SQL query and return results as dict
                          keyed on specified key (which should be unique)

    In addition the cursor can use the SQL API methods described below or
    access the underlying psycopg2 cursor (via the self.cursor attribute).

    The cursor methods are also available as standalone functions which
    run inside an implicit cursor object.

    SQL API
    -------

    The cursor class also provides a simple Python API for common SQL
    operations.  The basic methods provides are:

        select          - single table select (with corresponding select_one, select_dict methods)
        join            - two table join (with corresponding join_one, join_dict methods)
        insert          - SQL insert
        update          - SQL update
        delete          - SQL delete

    The methods can be parameterised to customise the associated query
    (see db module for detail):

        where           - 'where' clause as dict (column operators can be
                          specified using the colunm__operator format)

                          where = {'name':'abc','status__in':(1,2,3)}

        columns         - list of columns to be returned - these can
                          be real columns or expressions. If spefified
                          as a tuple the column is explicitly named
                          using the AS operator

                          columns = ('name',('status > 1','updated'))

        order           - sort order as list (use 'column__desc' to
                          reverse order)

                          order = ('name__desc',)

        limit           - row limit (int)

        offset          - offset (int)

        on              - join columns (as tuple)

        values          - insert data as dict

        returning       - columns to return (string)

    The methods are also available as standalone functions which create an
    implicit cursor object.

    Basic usage:

        >>> db.create_table('t1','id serial,name text,count int')
        >>> db.create_table('t2','id serial,t1_id int,value text')
        >>> db.log = sys.stdout
        >>> db.insert('t1',{'name':'abc','count':0},returning='id,name')
        INSERT INTO t1 (name) VALUES ('abc') RETURNING id,name
        [1, 'abc']
        >>> db.insert('t2',{'t1_id':1,'value':'t2'})
        INSERT INTO t2 (t1_id,value) VALUES (1,'t2')
        1
        >>> db.select('t1')
        SELECT * FROM t1
        [[1, 'abc', 0]]
        >>> db.select_one('t1',where={'name':'abc'},columns=('name','count'))
        SELECT name, count FROM t1 WHERE name = 'abc'
        ['abc', 0]
        >>> db.join(('t1','t2'),columns=('t1.id','t2.value'))
        SELECT t1.id, t2.value FROM t1 JOIN t2 ON t1.id = t2.t1_id
        [[1, 't2']]
        >>> db.insert('t1',{'name':'abc'},returning='id')
        INSERT INTO t1 (name) VALUES ('abc') RETURNING id
        [2]
        >>> db.update('t1',{'name':'xyz'},where={'name':'abc'})
        UPDATE t1 SET name = 'xyz' WHERE name = 'abc'
        2
        >>> db.update('t1',{'count__func':'count + 1'},where={'count__lt':10},returning="id,count")
        UPDATE t1 SET count = count + 1 WHERE count < 10 RETURNING id,count
        [[1, 1]]

    Prepared Statements
    -------------------

        Prepared statements can be created using the

            connection.prepare(stmt,params,name,call_type)

            stmt      : prepared statement (with parameters identified
                        in the statement using the psql $1,$2... notation)
            params    : list of optional parameter types (usually not
                        needed - infered by psql)
            name      : name for the prepared statement (usually
                        autogenerated)
            call_type : method used when instance called as method
                        (defaults to 'query')

        The constructor returns a PreparedStatement object which can be used
        instead of an sql statement in the connection.execute and
        connection.query_xxx methods.

        >>> p = db.prepare('UPDATE t1 SET name = $2 WHERE id = $1')
        PREPARE _pstmt_001  AS UPDATE t1 SET name = $2 WHERE id = $1
        >>> with db.cursor() as c:
        ...     c.execute(p,(1,'xxx'))
        EXECUTE _pstmt_001 (1,'xxx')

        The PreparedStatement object can also be called directly using the
        execute/query/query_one/query_dict methods. The instance is also
        directly callable using the method type identified in 'call_type'

        >>> p = db.prepare('UPDATE t1 SET name = $2 WHERE id = $1')
        PREPARE _pstmt_001  AS UPDATE t1 SET name = $2 WHERE id = $1
        >>> p.execute(1,'xxx')
        EXECUTE _pstmt_001 (1,'xxx')
        >>> p(1,'xxx')
        EXECUTE _pstmt_001 (1,'xxx')

    Logging
    -------

        To enable logging the connection.log attribute can be set to either an
        instance of logging.Logger or a file-like object (supporting the write
        method).

        The log message is generated using the self.logf function (called with
        the cursor object as a parameter). By default this just returns the
        query string however can be customised as needed. A cursor.timestamp
        attribute is available to allow execution time to be tracked.

        >>> db.log = sys.stdout
        >>> db.logf = lambda c : '[%f] %s' % (time.time() - c.timestamp,c.query)
        >>> db.query('SELECT * FROM t1')
        [0.000536] SELECT * FROM t1

    Changelog
    ---------

        *   0.1.0     03-06-2013  Initial import
        *   0.1.1     10-06-2013  Transaction context issues
        *   0.1.2     11-06-2013  ContextManager commit issues
        *   0.1.3     07-08-2013  ContextManager __exit__ fail on TypeError exception
        *   0.1.4     07-08-2013  ContextManager __exit__ fail on TypeError exception
        *   0.1.5     08-10-2013  - ThreadedConnectionPool fix when pool is exausted or max_con of Postgres is reached.
                                  - Created a param [pool_manager] in config_pool method. Params: SimpleConnectionPool,
                                  ThreadedConnectionPool. In Multthread enviroments must use ThreadedConnectionPool.
        *   0.1.6     14-10-2014  - Bugfix. Fix import of OperationalError. Avoid use protected member of psycopg2.
                                  - Change "from psycopg2._psycopg import OperationalError" to
                                    "from psycopg2 import OperationalError"
        *   0.1.11    04-09-2015  Non Threaded AutoCloseConnectionPool to use with pgpool
        *   0.1.12    04-09-2015  Deleted class AutoCloseConnectionPool.
                                  - Create configuration in env PYPGWRAP_CLOSE_CONNECTION_ON_EXIT to control
                                    when pypgwrap disable pool. When this env variable is True, all connections
                                    is discarded when execution is finished. No pooling is persisted. Util to use with
                                    PgPool or external pooling tools.
        *   0.1.13    17-09-2015  Fix use of ast.literal_eval to read Environment variable

    Author
    ------

        *   Erick Phillipe R. de Almeida (ephillipe@gmail.com)

    Master Repository/Issues
    ------------------------

        *   https://github.com/ephillipe/pypgwrap

    Credits
    ------------------------
        pypgwrap is inherited from pgwrap, an excelent wraper for Postgres but with lacks
        *   https://github.com/paulchakravarti/pgwrap

        Pooling is iherited from Psycopg2
        * https://github.com/psycopg/psycopg2/

    About me
    ------------------------
        *   http://about.me/erick.almeida
        *   http://erickalmeida.brandyourself.com



"""






import ast

__author__ = 'Erick Almeida'

import os
from collections import namedtuple
from psycopg2.extras import DictCursor, NamedTupleCursor
import psycopg2.extensions as _ext
from pool import SimpleConnectionPool, ThreadedConnectionPool
from cursor import cursor, PreparedStatement
from psycopg2 import OperationalError
from psycopg2.pool import PoolError

__connection_pool__ = ThreadedConnectionPool()


class SafeNamedTupleCursor(NamedTupleCursor):
    def _make_nt(self, namedtuple=namedtuple):
        return namedtuple("Record", [d[0] for d in self.description or ()], rename=True)


def config_pool(max_pool=5, pool_expiration=5, url=None, pool_manager=ThreadedConnectionPool):
    import urlparse

    params = urlparse.urlparse(url or os.environ.get('DATABASE_URL') or 'postgres://localhost/')
    global __connection_pool__
    __connection_pool__ = pool_manager()
    __connection_pool__.configure(expiration=pool_expiration,
                                  maxconn=max_pool,
                                  database=params.path[1:],
                                  user=params.username,
                                  password=params.password,
                                  host=params.hostname,
                                  port=params.port)


def get_pool():
    global __connection_pool__
    return __connection_pool__


class connection(object):
    def __init__(self, hstore=False, log=None, logf=None, default_cursor=DictCursor, key=None):
        self.pool = get_pool()
        self.key = key
        self.close_on_exit = ast.literal_eval(os.getenv('PYPGWRAP_CLOSE_CONNECTION_ON_EXIT', "False"))
        self.closed = False
        try:
            self.connection = self.pool.getconn(self.key)
        except (PoolError, OperationalError) as e:
            self.connection = None
            raise e
        self.hstore = hstore
        self.log = log
        self.logf = logf or (lambda cursor: cursor.query)
        self.default_cursor = default_cursor
        self.prepared_statement_id = 0

    def prepare(self, statement, params=None, name=None, call_type=None):
        """
            >>> db = connection()
            >>> p1 = db.prepare('SELECT name FROM doctest_t1 WHERE id = $1')
            >>> p2 = db.prepare('UPDATE doctest_t1 set name = $2 WHERE id = $1',('int','text'))
            >>> db.execute(p2,(1,'xxxxx'))
            1
            >>> db.query_one(p1,(1,))
            ['xxxxx']
            >>> db.execute(p2,(1,'aaaaa'))
            1
            >>> db.query_one(p1,(1,))
            ['aaaaa']
        """
        if not name:
            self.prepared_statement_id += 1
            name = '_pstmt_%03.3d' % self.prepared_statement_id
        if params:
            params = '(' + ','.join(params) + ')'
        else:
            params = ''
        with self.cursor() as c:
            c.execute('PREPARE %s %s AS %s' % (name, params, statement))
        if call_type is None:
            if statement.lower().startswith('select'):
                call_type = 'query'
            else:
                call_type = 'execute'
        return PreparedStatement(self, name, call_type)

    def cursor(self, cursor_factory=None):
        return cursor(self.connection,
                      cursor_factory or self.default_cursor,
                      self.hstore,
                      self.log,
                      self.logf)

    def __getattr__(self, name):
        def _wrapper(*args, **kwargs):
            with self.cursor() as c:
                return getattr(c, name)(*args, **kwargs)

        return _wrapper

    def commit(self, context_transaction=False):
        if self.connection:
            if self.key and not context_transaction:
                raise Exception(
                    'Connection was associated with Connection Context. Commits are not allowed. Use context_transaction if you want do it.')
            self.connection.commit()

    def rollback(self, context_transaction=False):
        if self.connection:
            if self.key and not context_transaction:
                raise Exception('Connection was associated with Connection Context. Rollbacks are not allowed.')
            self.connection.rollback()

    def close(self, context_transaction=False):
        if self.connection:
            if self.key and not context_transaction:
                raise Exception(
                    'Connection was associated with Connection Context. Commits are not allowed. Use context_transaction if you want do it.')
            self.pool.putconn(self.connection, close=self.close_on_exit)
            self.closed = True

    def __enter__(self, name=None):
        return self

    def __exit__(self, type, value, traceback):
        if not self.key:
            if not isinstance(value, Exception):
                self.commit()
            else:
                self.rollback()
            self.close()

    def __del__(self):
        if not self.key and self.connection and not self.closed:
            conn = self.connection
            self.pool.putconn(conn, close=self.close_on_exit)



__author__ = 'Erick Almeida'

import datetime
import psycopg2
import psycopg2.extensions as _ext
from psycopg2.pool import PoolError


class AbstractConnectionPool(object):
    """Generic key-based pooling code."""

    def __init__(self):
        self._pool = []
        self._used = {}
        self._rused = {}  # id(conn) -> key map
        self._tused = {}
        self._keys = 0
        self.closed = False

    def configure(self, expiration, maxconn, *args, **kwargs):
        """Initialize the connection pool.

        New 'minconn' connections are created immediately calling 'connfunc'
        with given parameters. The connection pool will support a maximum of
        about 'maxconn' connections.
        """
        self.expiration = expiration
        self.maxconn = maxconn

        self._args = args
        self._kwargs = kwargs

    def _connect(self, key=None):
        """Create a new connection and assign it to 'key' if not None."""
        conn = psycopg2.connect(*self._args, **self._kwargs)
        if key is not None:
            self._used[key] = conn
            self._rused[id(conn)] = key
            self._tused[id(conn)] = datetime.datetime.now()
        else:
            self._pool.append(conn)
        return conn

    def _disconnect(self, conn, remove_from_pool=False):
        if remove_from_pool and conn in self._pool:
            self._pool.remove(conn)
        conn.close()
        del self._tused[id(conn)]

    def _getkey(self):
        """Return a new unique key."""
        self._keys += 1
        return self._keys

    def _getconn(self, key=None):
        """Get a free connection and assign it to 'key' if not None."""
        if self.closed:
            raise PoolError("connection pool is closed")
        if key is None:
            key = self._getkey()
        if key in self._used:
            return self._used[key]

        if self._pool:
            self._used[key] = conn = self._pool.pop()
            self._rused[id(conn)] = key
            self._tused[id(conn)] = datetime.datetime.now()
            return conn
        else:
            if len(self._used) == self.maxconn:
                raise PoolError("connection pool exausted")
            return self._connect(key)

    def clear_expired_connections(self):
        now = datetime.datetime.now()
        expireds = []
        for item in self._pool:
            conn_time = self._tused[id(item)]
            minutes, seconds = divmod((now - conn_time).seconds, 60)
            if minutes >= self.expiration:
                expireds.append(item)
        for item in expireds:
            self._disconnect(item, True)

    def _putconn(self, conn, key=None, close=False):
        """Put away a connection."""
        if self.closed:
            raise PoolError("connection pool is closed")
        if key is None:
            key = self._rused.get(id(conn))
        if not key:
            raise PoolError("trying to put unkeyed [{key}] connection".format(key=key))

        if len(self._pool) < self.maxconn and not close:
            # Return the connection into a consistent state before putting
            # it back into the pool
            if not conn.closed:
                status = conn.get_transaction_status()
                if status == _ext.TRANSACTION_STATUS_UNKNOWN:
                    # server connection lost
                    self._disconnect(conn.close)
                elif status != _ext.TRANSACTION_STATUS_IDLE:
                    # connection in error or in transaction
                    conn.rollback()
                    self._pool.append(conn)
                else:
                    # regular idle connection
                    self._pool.append(conn)
                    # If the connection is closed, we just discard it.
        else:
            self._disconnect(conn)

        self.clear_expired_connections()

        # here we check for the presence of key because it can happen that a
        # thread tries to put back a connection after a call to close
        if not self.closed or key in self._used:
            del self._used[key]
            del self._rused[id(conn)]

    def _closeall(self):
        """Close all connections.

        Note that this can lead to some code fail badly when trying to use
        an already closed connection. If you call .closeall() make sure
        your code can deal with it.
        """
        if self.closed:
            raise PoolError("connection pool is closed")
        for conn in self._pool + list(self._used.values()):
            try:
                conn.close()
            except:
                pass
        self.closed = True

    def __del__(self):
        self._closeall()


class SimpleConnectionPool(AbstractConnectionPool):
    """A connection pool that can't be shared across different threads."""

    getconn = AbstractConnectionPool._getconn
    putconn = AbstractConnectionPool._putconn
    closeall = AbstractConnectionPool._closeall


class ThreadedConnectionPool(AbstractConnectionPool):
    """A connection pool that works with the threading module."""

    def __init__(self):
        """Initialize the threading lock."""
        import threading
        AbstractConnectionPool.__init__(self)
        self._lock = threading.Lock()

    def getconn(self, key=None):
        """Get a free connection and assign it to 'key' if not None."""
        self._lock.acquire()
        try:
            return self._getconn(key)
        finally:
            self._lock.release()

    def putconn(self, conn=None, key=None, close=False):
        """Put away an unused connection."""
        self._lock.acquire()
        try:
            self._putconn(conn, key, close)
        finally:
            self._lock.release()

    def closeall(self):
        """Close all connections (even the one currently in use.)"""
        self._lock.acquire()
        try:
            self._closeall()
        finally:
            self._lock.release()


import ast

__author__ = 'Erick Almeida'

import os
from connection import get_pool
import pypgwrap


class ContextManager(object):
    def __enter__(self):
        import uuid
        self._key = uuid.uuid4()
        self.pool = get_pool()
        self.close_on_exit = ast.literal_eval(os.getenv('PYPGWRAP_CLOSE_CONNECTION_ON_EXIT', "False"))
        return self

    def __exit__(self, type, value, traceback):
        conn = pypgwrap.connection(key=self._key)
        is_exception = isinstance(value, Exception) or ((not (type is None)) and issubclass(type, Exception))
        if not is_exception:
            conn.commit(context_transaction=True)
        else:
            conn.rollback(context_transaction=True)
        self.pool.putconn(conn.connection, key=self._key, close=self.close_on_exit)

    @property
    def key(self):
        return self._key



__author__ = 'Erick Almeida'

_operators = {'eq': '=',
              'lt': '<',
              'gt': '>',
              'ne': '!=',
              're': '~',
              'like': 'LIKE',
              'not_like': 'NOT LIKE',
}

_update_operators = {'': "%(field)s = %%(%(key)s)s",
                     'add': "%(field)s = %(field)s + %%(%(key)s)s",
                     'sub': "%(field)s = %(field)s - %%(%(key)s)s",
                     'append': "%(field)s = %(field)s || %%(%(key)s)s",
                     'func': "%(field)s = %(val)s",
}


def where(where):
    """
        Construct where clause from dict in format:

        eg. { 'key1'     : 'value1',
              'key2__op' : 'value1' }

            'key1 = %s AND key2 op %s' % (value1,value2) 

        'op' is looked up in '_operators' hash table which
        maps common opeartors (eg '__lt' to '<'). If the 
        operator is not found it is passed through directly
        allowing other operators to be specified directly.
    """
    if where:
        _where = []
        for f in where.keys():
            field, _, op = f.partition('__')
            _where.append('%s %s %%(%s)s' % (field, _operators.get(op, op) or '=', f))
        return ' WHERE ' + ' AND '.join(_where)
    else:
        return ''


def update(values):
    _update = []
    for k, v in values.items():
        f, _, op = k.partition('__')
        _update.append(_update_operators[op] %
                       {'key': k, 'val': v, 'field': f, 'op': op})
    return ','.join(_update)


def order(order):
    if order:
        _order = []
        for f in order:
            field, _, direction = f.partition('__')
            _order.append(field + (' DESC' if direction == 'desc' else ''))
        return ' ORDER BY ' + ', '.join(_order)
    else:
        return ''


def columns(columns):
    if columns:
        return ", ".join([(c if isinstance(c, (str, unicode))
                           else "%s AS %s" % c) for c in columns])
    else:
        return '*'


def on((t1, t2), on):
    if on:
        return "%s = %s" % on
    else:
        return "%s.id = %s.%s_id" % (t1, t2, t1)


def limit(limit):
    if limit:
        return ' LIMIT %d' % limit
    else:
        return ''


def offset(offset):
    if offset:
        return ' OFFSET %d' % offset
    else:
        return ''


def for_update(update):
    if update:
        return ' FOR UPDATE'
    else:
        return ''
      




#!/usr/bin/env python

try:
    from setuptools import setup, Command
except ImportError:
    from distutils.core import Command, setup

version = "0.1.13"
description = 'PostgreSQL database wrapper - provides wrapper over psycopg2 supporting a Python API for common sql ' \
              'functions, transaction and pooling'
long_description = file("README").read()


class GenerateReadme(Command):
    description = "Generates README file"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import pypgwrap
        import textwrap
        long_description_tmp = textwrap.dedent(pypgwrap.__doc__)
        open("README", "w").write(long_description_tmp)


setup(name='pypgwrap',
      version=version,
      description=description,
      long_description=long_description,
      platforms=["any"],
      author='Erick Almeida',
      author_email='ephillipe@gmail.com',
      maintainer="Erick Almeida",
      maintainer_email="ephillipe@gmail.com",
      url='https://github.com/ephillipe/pypgwrap',
      cmdclass={'readme': GenerateReadme},
      packages=['pypgwrap'],
      install_requires=['psycopg2'],
      license='BSD',
      classifiers=["Topic :: Database"]
)


__author__ = 'Erick Almeida'

import logging
import os
import time
import sqlop
import psycopg2


class cursor(object):
    def __init__(self, connection, cursor_factory, hstore, log, logf):
        self.connection = connection
        if cursor_factory:
            self.cursor_factory = cursor_factory
        else:
            self.cursor_factory = psycopg2.extensions.cursor
        self.hstore = hstore
        self.log = log
        self.logf = logf

    def _write_log(self, cursor):
        """
            >>> db = connection()
            >>> db.log = sys.stdout
            >>> _ = db.select('doctest_t1',columns=('name','count'),where={'active':True,'name__gt':'c'})
            SELECT name, count FROM doctest_t1 WHERE active = true AND name > 'c'
            >>> db.logf = lambda c : "--- %s ---" % c.query
            >>> _ = db.select('doctest_t1')
            --- SELECT * FROM doctest_t1 ---
        """
        msg = self.logf(cursor)
        if msg:
            if isinstance(self.log, logging.Logger):
                self.log.debug(msg)
            else:
                self.log.write(msg + os.linesep)

    def __enter__(self, name=None):
        """
            >>> db = connection()
            >>> with db.cursor() as c:
            ...     _ = c.insert('doctest_t1',values={'name':'xxx'})
            ...     _ = c.insert('doctest_t1',values={'name':'yyy'})
            ...     _ = c.insert('doctest_t1',values={'name':'zzz'})
            >>> db.select('doctest_t1',where={'name__~':'[xyz]+'},order=('name',),columns=('name',))
            [['xxx'], ['yyy'], ['zzz']]
            >>> db.delete('doctest_t1',where={'name__in':('xxx','yyy','zzz')},returning='name')
            [['xxx'], ['yyy'], ['zzz']]
            >>> with db.cursor() as c:
            ...     _ = c.insert('doctest_t1',values={'name':'xxx'})
            ...     c.commit()
            ...     _ = c.insert('doctest_t1',values={'name':'yyy'})
            ...     _ = c.insert('doctest_t1',values={'name':'zzz'})
            ...     c.rollback()
            >>> db.delete('doctest_t1',where={'name__in':('xxx','yyy','zzz')},returning='name')
            [['xxx']]
        """
        self.cursor = self.connection.cursor(name=name, cursor_factory=self.cursor_factory)
        if self.hstore:
            psycopg2.extras.register_hstore(self.cursor)
        return self

    def __exit__(self, type, value, traceback):
        self.cursor.close()

    def execute(self, sql, params=None):
        """
            >>> db = connection()
            >>> db.execute('select name,active FROM doctest_t1')
            10
        """
        if isinstance(sql, PreparedStatement):
            if params:
                sql = 'EXECUTE %s (%s)' % (sql.name, ','.join(['%s'] * len(params)))
            else:
                sql = 'EXECUTE %s' % sql.name
        if self.log and self.logf:
            try:
                self.cursor.timestamp = time.time()
                self.cursor.execute(sql, params)
                return self.cursor.rowcount
            finally:
                self._write_log(self.cursor)
        else:
            self.cursor.execute(sql, params)
            return self.cursor.rowcount

    def query(self, sql, params=None):
        """
            >>> db = connection()
            >>> r = db.query('select name,active FROM doctest_t1 ORDER BY name')
            >>> r[0]
            ['aaaaa', True]
            >>> len(r)
            10
        """
        self.execute(sql, params)
        return self.cursor.fetchall()

    def query_one(self, sql, params=None):
        """
            >>> db = connection()
            >>> db.query_one('select name,active FROM doctest_t1 WHERE name = %s',('aaaaa',))
            ['aaaaa', True]
        """
        self.execute(sql, params)
        return self.cursor.fetchone()

    def query_dict(self, sql, key, params=None):
        """
            >>> db = connection()
            >>> r = db.query_dict('select name,active FROM doctest_t1 ORDER BY name','name')
            >>> r['aaaaa']
            ['aaaaa', True]
            >>> sorted(r.keys())
            ['aaaaa', 'bbbbb', 'ccccc', 'ddddd', 'eeeee', 'fffff', 'ggggg', 'hhhhh', 'iiiii', 'jjjjj']
        """
        _d = {}
        for row in self.query(sql, params):
            _d[row[key]] = row
        return _d

    def _build_select(self, table, where, order, columns, limit, offset, update):
        return 'SELECT %s FROM %s' % (sqlop.columns(columns), table) \
               + sqlop.where(where) + sqlop.order(order) + sqlop.limit(limit) \
               + sqlop.offset(offset) + sqlop.for_update(update)

    def select(self, table, where=None, order=None, columns=None, limit=None, offset=None, update=False):
        """
            >>> db = connection()
            >>> db.select('doctest_t1') == db.query('SELECT * FROM doctest_t1')
            True
            >>> db.select('doctest_t1',columns=('name',),order=('name',),limit=2)
            [['aaaaa'], ['bbbbb']]
            >>> db.select('doctest_t1',where={'name__in':('aaaaa','bbbbb')},order=('name__desc',)) == \
                    db.query("SELECT * FROM doctest_t1 WHERE name IN ('aaaaa','bbbbb') ORDER BY name DESC")
            True
            >>> db.select_one('doctest_t1',columns=('name',),where={'name__in':('bbbbb',)})
            ['bbbbb']
        """
        return self.query(self._build_select(table, where, order, columns, limit, offset, update), where)

    def select_one(self, table, where=None, order=None, columns=None, limit=None, offset=None, update=False):
        """
            >>> db = connection()
            >>> db.select_one('doctest_t1',order=('name',),columns=('name',))
            ['aaaaa']
            >>> db.select_one('doctest_t1',order=('name',),columns=(('name','abcd'),))
            ['aaaaa']
        """
        return self.query_one(self._build_select(table, where, order, columns, limit, offset, update), where)

    def select_dict(self, table, key, where=None, order=None, columns=None, limit=None, offset=None, update=False):
        """
            >>> db = connection()
            >>> db.select_dict('doctest_t1','name',columns=('name',),order=('name',),limit=2)
            {'aaaaa': ['aaaaa'], 'bbbbb': ['bbbbb']}
        """
        return self.query_dict(self._build_select(table, where, order, columns, limit, offset, update), key, where)

    def _build_join(self, tables, where, on, order, columns, limit, offset):
        on = on or [None] * len(tables)
        return 'SELECT %s FROM %s ' % (sqlop.columns(columns), tables[0]) + \
               " ".join(['JOIN %s ON %s' % (tables[i], sqlop.on((tables[0], tables[i]), on[i - 1]))
                         for i in range(1, len(tables))]) + \
               sqlop.where(where) + sqlop.order(order) + sqlop.limit(limit) + sqlop.offset(offset)

    def join(self, tables, where=None, on=None, order=None, columns=None, limit=None, offset=None):
        """
            >>> db = connection()
            >>> db.join(('doctest_t1','doctest_t2'),columns=('name','value'),
            ...             where={'doctest_t1.name__in':('aaaaa','bbbbb','ccccc')},
            ...             order=('name',),limit=2)
            [['aaaaa', 'aa'], ['bbbbb', 'bb']]
            >>> db.join(('doctest_t1','doctest_t2'),on=[('doctest_t1.id','doctest_t2.doctest_t1_id')]) \
                            == db.join(('doctest_t1','doctest_t2'))
            True
        """
        return self.query(self._build_join(tables, where, on, order, columns, limit, offset), where)

    def join_one(self, tables, where=None, on=None, order=None, columns=None, limit=None, offset=None):
        """
            >>> db = connection()
            >>> db.join_one(('doctest_t1','doctest_t2'),columns=('name','value'),where={'name':'aaaaa'})
            ['aaaaa', 'aa']
        """
        return self.query_one(self._build_join(tables, where, on, order, columns, limit, offset), where)

    def join_dict(self, tables, key, where=None, on=None, order=None, columns=None, limit=None, offset=None):
        """
            >>> db = connection()
            >>> db.join_dict(('doctest_t1','doctest_t2'),'name',columns=('name','value'),
            ...               where={'doctest_t1.name__in':('aaaaa','bbbbb','ccccc')},
            ...               order=('name',),limit=2)
            {'aaaaa': ['aaaaa', 'aa'], 'bbbbb': ['bbbbb', 'bb']}
        """
        return self.query_dict(self._build_join(tables, where, on, order, columns, limit, offset), key, where)

    def insert(self, table, values, returning=None):
        """
            >>> db = connection()
            >>> db.insert('doctest_t1',{'name':'xxx'})
            1
            >>> db.insert('doctest_t1',values={'name':'yyy'},returning='name')
            ['yyy']
            >>> db.insert('doctest_t1',values={'name':'zzz'},returning='name')
            ['zzz']
            >>> db.select('doctest_t1',where={'name__~':'[xyz]+'},order=('name',),columns=('name',))
            [['xxx'], ['yyy'], ['zzz']]
            >>> db.delete('doctest_t1',where={'name__in':('xxx','yyy','zzz')})
            3
        """
        _values = ['%%(%s)s' % v for v in values.keys()]
        sql = 'INSERT INTO %s (%s) VALUES (%s)' % (table, ','.join(values.keys()), ','.join(_values))
        if returning:
            sql += ' RETURNING %s' % returning
            return self.query_one(sql, values)
        else:
            return self.execute(sql, values)

    def delete(self, table, where=None, returning=None):
        """
            >>> db = connection()
            >>> db.insert('doctest_t1',{'name':'xxx'})
            1
            >>> db.insert('doctest_t1',{'name':'xxx'})
            1
            >>> db.delete('doctest_t1',where={'name':'xxx'},returning='name')
            [['xxx'], ['xxx']]
        """
        sql = 'DELETE FROM %s' % table + sqlop.where(where)
        if returning:
            sql += ' RETURNING %s' % returning
            return self.query(sql, where)
        else:
            return self.execute(sql, where)

    def update(self, table, values, where=None, returning=None):
        """
            >>> db = connection()
            >>> db.insert('doctest_t1',{'name':'xxx'})
            1
            >>> db.update('doctest_t1',{'name':'yyy','active':False},{'name':'xxx'})
            1
            >>> db.update('doctest_t1',values={'count__add':1},where={'name':'yyy'},returning='count')
            [[1]]
            >>> db.update('doctest_t1',values={'count__add':1},where={'name':'yyy'},returning='count')
            [[2]]
            >>> db.update('doctest_t1',values={'count__func':'floor(pi()*count)'},where={'name':'yyy'},returning='count')
            [[6]]
            >>> db.update('doctest_t1',values={'count__sub':6},where={'name':'yyy'},returning='count')
            [[0]]
            >>> db.delete('doctest_t1',{'name':'yyy'})
            1
        """
        sql = 'UPDATE %s SET %s' % (table, sqlop.update(values))
        sql = self.cursor.mogrify(sql, values)
        if where:
            sql += self.cursor.mogrify(sqlop.where(where), where)
        if returning:
            sql += ' RETURNING %s' % returning
            return self.query(sql)
        else:
            return self.execute(sql)

    def check_table(self, t):
        """
            >>> db = connection()
            >>> db.check_table('doctest_t1')
            True
            >>> db.check_table('nonexistent')
            False
        """
        _sql = 'SELECT tablename FROM pg_tables WHERE schemaname=%s and tablename=%s'
        return self.query_one(_sql, ('public', t)) is not None

    def drop_table(self, t):
        """
            >>> db = connection()
            >>> db.create_table('doctest_t3','''id SERIAL PRIMARY KEY, name TEXT''')
            >>> db.check_table('doctest_t3')
            True
            >>> db.drop_table('doctest_t3');
            >>> db.check_table('doctest_t3')
            False
        """
        self.execute('DROP TABLE IF EXISTS %s CASCADE' % t)

    def create_table(self, name, schema):
        """
            >>> db = connection()
            >>> db.create_table('doctest_t3','''id SERIAL PRIMARY KEY, name TEXT''')
            >>> db.check_table('doctest_t3')
            True
            >>> db.drop_table('doctest_t3');
            >>> db.check_table('doctest_t3')
            False
        """
        if not self.check_table(name):
            self.execute('CREATE TABLE %s (%s)' % (name, schema))


class PreparedStatement(object):
    def __init__(self, connection, name, call_type='query'):
        self.connection = connection
        self.name = name
        self.call_type = call_type

    def deallocate(self):
        self.connection.execute('DEALLOCATE %s' % self.name)

    def execute(self, *params):
        return self.connection.execute(self, params)

    def query(self, *params):
        return self.connection.query(self, params)

    def query_one(self, *params):
        return self.connection.query_one(self, params)

    def query_dict(self, key, *params):
        return self.connection.query_dict(self, key, params)

    def __call__(self, *params):
        return getattr(self.connection, self.call_type)(self, params)